(     Title:  Paartial Reference Implementation
              for IEEE 754-2008 Floating Point
       File:  ieeefp-ref.fs
  Test file:  ieeefp-ref-test.fs
    Version:  0.5.0
    Revised:  October 16, 2020
     Author:  David N. Williams
    License:  Creative Commons CC0 1.0 Universal

The revision date may reflect cosmetic changes not logged below.

Version 0.5.0

11Sep20 * Started.
        * Added bit manipulation constants, RAW-CONSTANT, RAW-DROP, RAW=.
        * Added classification queries. 
12Sep20 * Added the 10 raw constants with positive sign, |RAW|.
        * Rewrote classification queries.
15Sep20 * Started comparisons with T-F<.
18Sep20 * Finished T-F<, a serious effort.
19Seo20 * Added:  T-F>  T-F=  T-F<=  T-F>=
30Sep20 * Made into a module named IEEEFP-REF.
        * Removed T- prefixes.
 1Oct20 * Added:  FDATUM=  FABSDIFF<  F~  FABS
 2Oct20 * Added:  FNEGATE  FCOPYSIGN  FMAX  FMIN
 3Oct20 * Rewrote analysis of FMAX and FMIN, not easy.
 9Oct20 * Added:  RAW-1+  RAW-1-  FNEXTUP  FNEXTDOWN  -MAX-N-RAW
10Oct20 * Added 80-bit gap handling to FNEXTUP.
11Oct20 * Added:  FSNAN?  FQNAN?
12Oct20 * Moved raw constants and RAW-DROP into rawfloat.fs.
15Oct20 * Replaced 32-BITS/CELL? by bits/cell expression.

\ *** OVERVIEW

This library works with 32|64-bit cells, and a total big|little-endian
binary floating-point format of the same endianness as that for cpu
integers.

It provides reference implementations of selected parts of "Proposal for
an Optional IEEE 754, Binary Floating-Point Word Set", draft version
0.5.5, August 31, 2009.  A copy can be found at:

  http://www.umich.edu/~williams/archive/forth/ieeefp-drafts/

Further development of the proosal is being taken up by Krisnha Myneni:

  https://forth-standard.org/~KrishnaMyneni

This library loads rawfloat.fs, a small libary based on a multi-cell,
cell-wise big-endian representation on the data stack of the Forth
system default binary floating-point format, called the raw-float
format, to get convenient access its bits.  The rawfloat library
requires that the default format have a width of 32, 64, 80, or 128
bits, corresponding to an IEEE 754-2008 single, double, [intel] extended
double, or quad format.

The library has not been tested with single [FIXME!!] or quad floats,
nor on a big-endian system.

Much of the action has to do with the extended ieee data types, signed
zero, signed infinity, and the various nans.  Because the reference
implementation treats these quantities through the raw-float format, it
is expected to work on Forth systems which do not explicitly implement
extended data.

There is, however, an assumption that the cpu has IEEE 754 floating
point, and that the Forth system has an implementation of the optional
Floating-Point word set for which F@ amd F! properly handle extended
data. 

The test file ieeefp-ext-test.fs assumes that this has been verfied by
testing the rawfloat library, so it freely uses conversions between the
raw and default formats based on F@ and F!.  It double checks that with
a few of its own tests.

MODULARITY:  This library loads the modules.fs library [km, dnw], which
allows the selective configuration of tests to use either reference or
native host implementations.  That includes the use of F~ in the
floating-point part of the tester itself.

NOTATION:  Since Forth floating point is beging extended to include IEEE
754-2008 special values, in floating-point stack comments we use the
Forth r for finite real values including signed zero, x for extended
real values including signed infinity, and f for a general ieee value
including quiet or signaling signed nans.

For floating-point quantities, we use "=", "<", and ">" to mean IEEE 754
comparisons.  When we want to say that two fp quantities have identical
encodings, we try to use "is" instead of "=".  Thus, +0 = -0 is true,
but +0 is -0 is false.

For raw quantities and their individual cells, we use "=", "<", and ">"
in the conventional sense.
)

\ *** WORDS
(
See rawfloat.fs for raw multicell definitions and basic manipulations.
)

\ CONSTANTS
\   signbit-raw.hi  +inf-raw.hi  +max-subn-raw.hi  +min-qnan-raw.hi+
\   [extended double only]
\   intbit-raw.nx  +min-qnan-raw.nx

\ RAW MANIPULATION
\   raw=  |raw|  raw-negate
\   raw-1+  raw-1-

\ CLASSIFICATION
\   FSIGNBIT  F0=  FINFINITE?  FNAN?  FSNAN?  FQNAN?  FFINITE?
\   FSUBNORMAL?  FNORMAL?

\ SIGN MANIPULATION
\   FNEGATE  FABS  FCOOPYSIGN

\ COMPARISON
\   F<  F>  F=  F<=  F>=
\   FDATUM=  FABSDIFF<  F~

\ DATA MANIPULATION
\   FMAX  FMIN  FNEXTUP  FNEXTDOWN
\   [not implemented yet]
\   FSCALBN  FLOGB  FREMAINDER
\   FCEIL  FLOOR  FROUND  FTRUNC  FNEARBYINT

[UNDEFINED] REQUIRE [IF] INCLUDE required-ref.fs [THEN]

REQUIRE modules.fs
REQUIRE rawfloat.fs

[UNDEFINED] (f: [IF] : (f: postpone ( ; IMMEDIATE [THEN]

[UNDEFINED] -FROT [IF]
: -FROT  (f: f1 f2 f3 -- f3 f1 f2 )  frot frot ; [THEN]
[UNDEFINED] F2DUP  [IF]
: F2DUP  (f: f1 f2 -- f1 f2 f1 f2 )  fover fover ; [THEN]
[UNDEFINED] F2DROP [IF]
: F2DROP (f: f1 f2 -- f1 f2 f1 f2 )  fdrop fdrop ; [THEN]
[UNDEFINED] FNIP [IF]
: FNIP  (f: f1 f2 -- f2 )  fswap fdrop ; [THEN]

MODULE: IEEEFP-REF

\ Debugging
: .#  ( n -- )  cr ." #" . .s cr ; 

\ *** CONSTANTS

1
bits/float #32 =  [IF] #31                                        [THEN]
bits/float #64 =  [IF] bits/cell #32 = [IF] #31 [ELSE] #63 [THEN] [THEN]
bits/float #80 =  [IF] #15                                        [THEN]
bits/float #128 = [IF] bits/cell #32 = [IF] #31 [ELSE] #63 [THEN] [THEN]
lshift CONSTANT signbit-raw.hi  \ .hi means most significant cell

bits/float #80 = [IF]
  bits/cell #32 = [IF] $80000000 [ELSE] $8000000000000000 [THEN]
  CONSTANT intbit-raw.nx   \ .nx means next most significant cell
[THEN]

+inf-raw      dup CONSTANT +inf-raw.hi      raw-drop 
+max-subn-raw dup CONSTANT +max-subn-raw.hi raw-drop
+min-qnan-raw dup CONSTANT +min-qnan-raw.hi
bits/float #80 = [IF]
  over CONSTANT +min-qnan-raw.nx [THEN] raw-drop

\ *** RAW OPERATIONS

: raw=  ( raw1 raw2 -- [=]? )  \ raw data exactly equal
  [ bits/float #32 = ]  [IF] = [THEN]

  [ bits/float #64 = ]  [IF] [ bits/cell #32 = ]
    [IF] d= [ELSE] = [THEN] [THEN]

  [ bits/float #80 = ]  [IF] [ bits/cell #32 = ]
    [IF] d= >r = r> and [ELSE] d= [THEN] [THEN]

  [ bits/float #128 = ] [IF] [ bits/cell #32 = ]
    [IF] d= >r d= r> and [ELSE] d= [THEN] [THEN] ;

: |raw| ( raw -- |raw| )  \ clear the raw sign bit
  [ signbit-raw.hi invert ] literal and ;

: raw-negate  ( raw -- -raw )  signbit-raw.hi xor ;

bits/cell #32 = [IF]  \ !!UNTESTED!!
: [UD1+]-CARRY  ( ud -- ud' carry )
  dup $80000000 and >r 1 0
  d+ dup $80000000 and 0= r> and [ bits/cell 1- ] literal rshift ;

: UT1+  ( ut -- ut+1 )  >r [ud1+]-carry r> + ;

: [UD1-]-BORROW  ( ud -- ud' borrow )
  dup $80000000 and >r 1 0
  d- dup $80000000 and 0= r> and negate ;

: UT1-  ( ut -- ut+1 )  >r [ud1-]-borrow r> + ;
[THEN]  \ !!UNTESTED!!

: raw-1+  ( raw -- raw+1 )
  [ bits/float #32 = ] [IF] 1+ [THEN]

  [ bits/float #64 = ] [IF]
    [ bits/cell #32 = ] [IF] 1 0 d+ [ELSE] 1+ [THEN] [THEN]

  [ bits/float #80 = ] [IF]
    [ bits/cell #32 = ] [IF] ut1+ [ELSE] 1 0 d+ [THEN] [THEN] ;

: raw-1-  ( raw -- raw-1 )
  [ bits/float #32 = ] [IF] 1- [THEN]

  [ bits/float #64 = ] [IF]
    [ bits/cell #32 = ] [IF] 1 0 d- [ELSE] 1- [THEN] [THEN]

  [ bits/float #80 = ] [IF]
    [ bits/cell #32 = ] [IF] ut1- [ELSE] 1 0 d- [THEN] [THEN] ;

BEGIN-MODULE
PUBLIC:

\ *** CLASSIFICATION

: FSIGNBIT  (f: r -- ) ( -- [-]? )
  f>raw dup >r raw-drop r> signbit-raw.hi and 0<> ;

: F0=  (f: r -- ) ( -- [0]? )
(
The IEEE 754 comparison ignores the sign of zero.
)
  f>raw |raw| +0-raw raw= ;

: FINFINITE?  (f: r -- ) ( -- [+|-inf]? )  f>raw |raw| +inf-raw raw= ;

: FNAN?  (f: r -- ) ( -- [nan]? )
(
True if r is not infinite, and the magnitude of the most sig raw cell is
>= the most sig cell of +inf-raw.  Algorithm from inspection of bit
patterns.
)
  fdup FINFINITE? 0= ( [not.inf]?)
  f>raw |raw| dup >r raw-drop r> +inf-raw.hi >= and ;

: FSNAN?  (f: f -- ) ( -- [snan]? )
  fdup FINFINITE? 0= ( [not.inf]?)
[ bits/float #80 = ] [IF]
  f>raw |raw| dup +inf-raw.hi = ( [snan|qnqn]?) >r ( [not.inf]? |f|.raw)
  over ( raw.nx) >r raw-drop ( [not.inf]?)
  +min-qnan-raw.nx r> ( raw.nx ) u> and r> and ;
[ELSE]
  f>raw |raw| dup >r raw-drop r@ +inf-raw.hi >= and
  +min-qnan-raw.hi r> u> and ;
[THEN]

\ Not required by IEEE 754-2008.
: FQNAN?  (f: f -- ) ( -- [qnan]? )  fdup FNAN? FSNAN? 0= and ;

: FFINITE?  (f: r -- ) ( -- [not-inf & not-nan]? )
  fdup FINFINITE? FNAN? or 0= ;

: FSUBNORMAL?  (f: r  -- ) ( -- [subnormal]? )
  fdup F0= 0= fdup FFINITE? and
  f>raw dup >r raw-drop r> |raw| +max-subn-raw.hi <= and ;

: FNORMAL?  (f: r -- ) ( -- [normal]? )
  fdup F0= 0= fdup FFINITE? and FSUBNORMAL? 0= and ;

\ *** SIGN MANIPULATION

: FNEGATE  (f: f -- -f )  f>raw raw-negate raw>f ;
: FABS     (f: f -- |f|)  f>raw |raw| raw>f ;

: FCOPYSIGN (f: f1 f2 -- f3 )
  f>raw dup >r raw-drop f>raw 
  [ signbit-raw.hi invert ] literal and r> signbit-raw.hi and or raw>f ; 

\ *** COMPARISON

: F<  ( f: f1 f2 -- ) ( -- [f1<f2]? )
(
If f1, f2 are both finite, use the host F< and exit.

If either f1 or f2 is a nan, leave false and exit.

If [x1, x2] = [-inf, -inf] or [x1, x2] = [+inf, +inf], leave false and
exit.

Apply signbit[x1] or not.signbit[x2] as appropriate from the following
truth tables and exit.

      x1.is.inf       [x1<x2]?    signbit[x1]
[1.a] +inf < +|x2|    false       false
[1.b] +inf < -|x2|    false       false
[1.c] -inf < +|x2|    true        true
[1.d] -inf < -|r2|    true        true
[1.e] -inf < -inf     false       true

Eliminating [1.e] means [x1<x2]? = signbit[x1] when x1.is.inf.
 
      x2.is.inf       [x1<x2]?    signbit[x2]
[2.a] +inf  < +inf    false       false
[2.b] +|r1| < +inf    true        false
[2.c] -|x1| < +inf    true        false
[2.d] +|x1| < -inf    false       true
[2.e] -|x1| < -inf    false       true

Eliminating [2.a] means [x1<x2]? = not.signbit[x2] when x2.is.inf.
)
\ Eliminate both finite, the most probable case.
  f2dup (f: f2) FFINITE? (f: f1) FFINITE?
  (f: f1 f2) ( [f2.fin]? [f1.fin]?) 2dup and
  IF 2drop f< EXIT THEN
\ At least one is not finite.

\ Below here, f1.is.not.fin or f2.is.not.fin.

\ Eliminate any nan.
  (f: f1 f2) ( [f2.fin]? [f1.fin]?)
  f2dup (f: f2) FNAN? (f: f1) FNAN?
  ( [f2.nan]? [f1.nan]?) or
  IF  \ f1.is.nan or f2.is.nan
    2drop f2drop false EXIT THEN
\ Nans are eliminated.

\ Below here, f1.is.not.nan and f2.is.not.nan means both f1 and f2 are
\ extended reals, and since at least one is not finite, at least one is
\ infinite.

\ Eliminate (x1, x2) = (-inf, -inf) or (+inf, +inf). 
  (f: x1 x2) ( [x2.fin]? [x1.fin]?) 0= ( [x1.inf]?)
  IF  \ x1.is.inf
    ( [x2.fin]?) 0= ( [x2.inf]?)
    f2dup (f: x2) FSIGNBIT (f: x1) FSIGNBIT xor 0= ( [same.sign]?)
    ( [x2.inf]?) and IF f2drop false EXIT THEN   
\ Both [1.e] and [2.a] are eliminated.

\ Because x1.is.inf, use [1.a-d].
   (f: x1 x2) fdrop FSIGNBIT
  ELSE  \ x1.is.fin

\ Because x2.is.inf, use [2.b-e].
    ( [x2.inf]?) drop (f: x1 x2) FSIGNBIT 0= fdrop
  THEN ;

: F>  (f: f1 f2 -- ) ( -- [f1>f2]? )  fswap F< ;

: F=  (f: f1 f2 -- ) ( -- [f1=f2]? )
  f2dup F0= F0= and    IF f2drop true  EXIT THEN
  f2dup FNAN? FNAN? or IF f2drop false EXIT THEN
  f>raw f>raw ( raw2 raw1) raw= ;  \ RAW= is symmetric

: F<=  (f: f1 f2 -- ) ( -- [f1<=f2]? )  f2dup F< F= or ;
: F>=  (f: f1 f2 -- ) ( -- [f1>=f2]? )  f2dup F> F= or ;

: FDATUM=  (f: f1 f2 -- ) ( -- [f1.dat=f2.dat]? )  f>raw f>raw raw= ;

: FABSDIFF<  (f: f1 f2 f3 -- ) ( -- flag )  -frot f- FABS F> ;

: F~  (f: f1 f2 f3 -- )  ( -- flag )
  fdup F0=         IF fdrop FDATUM= EXIT THEN
  fdup FNAN?       IF f2drop fdrop false EXIT THEN
  fdup FSIGNBIT 0= IF FABSDIFF< EXIT THEN
  \ From here the nan's will propapate and be handled correctly by F<.
  FNEGATE -frot             (f: |f3| f1 f2)
  f2dup f- FABS             (f: |f3| f1 f2 |f1-f2|)
  -frot FABS fswap FABS f+  (f: |f3| |f1-f2| |f2|+|f1|)
  frot                      (f: |f1-f2| |f2|+|f1| |f3|)
  f* F< ;

\ *** DATA MANIPULATION

: FMAX  (f: f1 f2 -- x | nan )
(
The following is based on IEEE 754-2008 Sec. 5.3.1.

IF f1 < f2, then return f2.
IF f1 > f2, then return f1.

In either of these cases, [f1,f2] is [x1,x2], and it is impossible that
both values are signed zeros.

If only one of f1,f2 is a nan, return the other, which is an extended
real.  [This is unlike the usual operations on nans that produce an f
result.]

Otherwise return either.  This has two subcases:

a. [f1,f2] is [x1,x2] with x1 = x2.  Hence if [x1,x2] is [+0,-0] or
   [-0,+0], the return can be either +0 or -0.

b. [f1,f2] is [nan1,nan2], and either can be returned.

The IEEE spec mentions that x returns are "canonicalized", which is
irrelevant for binary formats.

It also mentions that with signaling nan input, the result must obey
Sec. 6.2.  This implementation is possibly noncompliant in that it
ignores any distinction between quiet and signaling.
)
  f2dup F<   IF fnip  (f: x2) EXIT THEN
  f2dup F>   IF fdrop (f: x1) EXIT THEN
  fdup FNAN? IF fdrop (f: f1) ELSE fnip (f: f2) THEN ;

: FMIN  (f: f1 f2 -- min[x1,x2] | nan )
(
The specification is the same as that for FMAX, but with < and >
interchanged.
)
  f2dup F>   IF fnip  (f: x2) EXIT THEN
  f2dup F<   IF fdrop (f: x1) EXIT THEN
  fdup FNAN? IF fdrop (f: f1) ELSE fnip (f: f2) THEN ;

: FNEXTUP  (f: f1 -- f2 )
(
Paraphrased from IEEE 754-2008 Standard Sec. 3.5.1:
a. if f1 is the negative number of least magnitude, then f2 is -0;  
b. nextUp[±0] is the positive nonzero number of least magnitude;
c. nextUp[+∞] is +∞;
d. nextUp[−∞] is the finite negative number of largest magnitude;
e. when f1 is NaN, then the result is according to Sec. 6.2;
f. nextUp[f1] does not signal an exception except for sNaNs.

The interpretation of Section 6.2 is not totally clear to us.  We take
it to mean that a quiet nan "should" be preserved, and that a signaling
nan "shall" be returned as quiet under default exception handling.

The basic algorithm is by Andrew Haley.
)
  fdup FINFINITE?
  IF FSIGNBIT IF -max-n-raw ELSE +inf-raw THEN raw>f EXIT THEN 

  \ finite or nan
  0.0e f+  \ convert -0 into +0, sNaN into qNaN
  f>raw dup signbit-raw.hi and ( signbit) >r
  |raw| dup +inf-raw.hi <
  IF  \ finite

[ bits/float #80 = ] [IF]

    r@ ( signbit)
    IF \ minus
      over intbit-raw.nx and
      IF \ normal
        raw-1- over intbit-raw.nx and 0=
        IF \ possible gap
          ( exp) 1-
          dup ( exp) 0<> ( gap?)
          IF \ output is normal, restore intbit to skip gap
            swap intbit-raw.nx or ( skip_gap) swap
          THEN
          \ if no gap, output is subnormal and correct
        THEN
      ELSE \ subnormal, not zero
        raw-1-
      THEN

    ELSE \ plus
      over intbit-raw.nx and
      IF  \ normal
        raw-1+ swap intbit-raw.nx or ( skip_any_gap) swap
      ELSE \ subnormal
        raw-1+ over intbit-raw.nx and IF drop 1 ( min-n) THEN
      THEN 
    THEN 

[ELSE] \ bits/float #80 <>

    r@ ( signbit) IF raw-1- ELSE raw-1+ THEN

[THEN]

    r> ( signbit) or raw>f EXIT
  THEN

  \ nan
  r> or raw>f ;

: FNEXTDOWN  (f: f1 -- f2 )  FNEGATE FNEXTUP FNEGATE ;

END-MODULE
